%%% -*- erlang-indent-level: 2 -*-
%%%-------------------------------------------------------------------
%%% File    : dialyzer_run.erl
%%% Authors : Richard Carlsson <richardc@csd.uu.se>
%%% Description : Provides a better way to start Dialyzer from a script.
%%%
%%% Created : 17 Oct 2004 by Richard Carlsson <richardc@csd.uu.se>
%%%-------------------------------------------------------------------
-module(dialyzer_run).

-include("dialyzer.hrl").
-include("hipe_icode_type.hrl").

-export([check_plt/0, gui/1, cli/1, build_options/1]).

%% Interfaces

check_plt() ->
  F = fun () -> dialyzer_cl:check_init_plt() end,
  run(F).

gui(Args) ->
  F = fun () ->
	  case parse_args(Args) of
	    [Opts] when is_list(Opts) ->
	      OptsRecord = build_options(Opts),
	      dialyzer_gui:start(OptsRecord);
	    Other ->
	      invalid_args("dialyzer_run:gui/1", Other)
	  end
      end,
  run(F).

cli(Args) ->
  F = fun () ->
	  case parse_args(Args) of
	    [Opts] when is_list(Opts) ->
	      %erlang:display(Opts),
	      OptsRecord = build_options(Opts),
	      dialyzer_cl:start(OptsRecord);
	    Other ->
	      invalid_args("dialyzer_run:cli/1", Other)
	  end
      end,
  run(F).


%% Machinery

run(F) ->
  wait_init(),
  case catch {ok, F()} of
    {ok, _} ->
      ok;
    {'EXIT', normal} ->
      ok;
    {'EXIT', E} ->
      report("terminated abnormally: ~P.", [E, 10]),
      error;
    Thrown ->
      report("throw without catch: ~P.", [Thrown, 15]),
      error
  end.

wait_init() ->
  case erlang:whereis(code_server) of
    undefined ->
      erlang:yield(),
      wait_init();
    _ ->
      ok
  end.

invalid_args(Where, Args) ->
  report("invalid arguments to ~s: ~w.", [Where, Args]),
  exit(error).

parse_args([A | As]) when atom(A) ->
  [parse_arg(atom_to_list(A)) | parse_args(As)];
parse_args([A | As]) ->
  [parse_arg(A) | parse_args(As)];
parse_args([]) ->
  [].

parse_arg(A) ->
  case catch {ok, parse(A)} of
    {ok, Expr} ->
      case catch erl_parse:normalise(Expr) of
	{'EXIT', _} ->
	  report("bad argument: '~s':", [A]),
	  exit(error);
	Term ->
	  Term
      end;
    {error, S} ->
      report("error parsing argument '~s'", [A]),
      report(S, []),
      exit(error)
  end.

parse(S) ->
  case erl_scan:string(S ++ ".", 1) of
    {ok, Ts, _} ->
      case erl_parse:parse_exprs(Ts) of
	{ok, [Expr]} ->
	  Expr;
	{error, E} ->
	  parse_error(E)
      end;
    {error, E, _} ->
      parse_error(E)
  end.

parse_error({_L, M, D}) ->
  throw({error,M:format_error(D)});
parse_error(E) ->
  %% Just in case we get some other error descriptor.
  S = io_lib:fwrite("error parsing expression: ~P.",[E,15]),
  throw({error,S}).

report(S, As) ->
  io:fwrite(S, As),
  io:nl().

build_options(Opts) ->
  DefaultWarns = [?WARN_RETURN_NO_RETURN,
		  ?WARN_NOT_CALLED,
		  ?WARN_NON_PROPER_LIST,
		  ?WARN_TUPLE_AS_FUN,
		  ?WARN_FUN_APP,
		  ?WARN_MATCHING,
		  ?WARN_COMP,
		  ?WARN_GUARDS,
		  ?WARN_OLD_BEAM,
		  ?WARN_FAILING_CALL],
  build_options(Opts, #options{legal_warnings=ordsets:from_list(DefaultWarns)}).

build_options([Term={OptionName,Value}|Rest], Options) ->
  case OptionName of
    files ->
      build_options(Rest, Options#options{files=Value});
    defines ->
      build_options(Rest, Options#options{defines=ordsets:from_list(Value)});
    from ->
      build_options(Rest, Options#options{from=Value});
    include_dirs ->
      build_options(Rest, Options#options{include_dirs=Value});
    output_file ->
      build_options(Rest, Options#options{output_file=Value});
    warnings ->
      NewWarnings = build_warnings(Value, Options#options.legal_warnings),
      build_options(Rest, Options#options{legal_warnings=NewWarnings});
    _ ->
      bad_option(Term)
  end;
build_options([Term|_Rest], _Options) ->
  bad_option(Term);
build_options([], Options) ->
  Options.

bad_option(Term) ->
  report("error building dialyzer options: ~P.\n",[Term,15]),
  exit(error).


build_warnings([Opt|Left], Warnings) ->
  NewWarnings =
    case Opt of
      no_return ->
	ordsets:del_element(?WARN_RETURN_NO_RETURN, Warnings);
      no_unused ->
	ordsets:del_element(?WARN_NOT_CALLED, Warnings);
      no_improper_list ->
	ordsets:del_element(?WARN_NON_PROPER_LIST, Warnings);
      no_tuple_as_fun ->
	ordsets:del_element(?WARN_TUPLE_AS_FUN, Warnings);
      no_fun_app ->
	ordsets:del_element(?WARN_FUN_APP, Warnings);
      no_match ->
	ordsets:del_element(?WARN_MATCHING, Warnings);
      no_comp ->
	ordsets:del_element(?WARN_COMP, Warnings);
      no_guards ->
	ordsets:del_element(?WARN_GUARDS, Warnings);
      no_unsafe_beam ->
	ordsets:del_element(?WARN_OLD_BEAM, Warnings);
      no_fail_call ->
	ordsets:del_element(?WARN_FAILING_CALL, Warnings);
      error_handling ->
	ordsets:add_element(?WARN_RETURN_ONLY_EXIT, Warnings);
      Other ->
	bad_option(Other)
    end,
  build_warnings(Left, NewWarnings);
build_warnings([], Warnings) ->
  Warnings.
